// bmpconv.cpp: dfinit le point d'entre pour l'application console.
//

#include "stdafx.h"
#include <stdlib.h>
#include <string.h>
#include "bmpconv.h"

#pragma warning(disable:4996)

int ConvertBMPFile(const char*, int flags);
void Convert(const char* name, char* buffer, int flags);
void ConvertToPal8(const char* name, char* buffer, int flags);
void Convert24ToRaw16(const char* name, char* buffer);
char* DestFileNamePal8(const char* filename);
char* DestFileNameRaw8(const char* filename);
char* DestFileNameRaw16(const char* filename);
void WriteFile(const char* filename, char* buffer, int size);

int main(int argc, char* argv[])
{
	if(argc>1)
	{
		int flags = 0;
		char* filename = NULL;
		for(int i = 1; i<argc; i++)
		{
			printf(argv[i]);
			if(strcmp(argv[i], "-p") == 0)
			{				
				flags = REMOVE_PALETTE;
			}
			else
			{
				filename = argv[i];
			}
			
		}
		printf("\n");
		return ConvertBMPFile(filename, flags);
	}
	return 0;
}

int ConvertBMPFile(const char* filename, int flags)
{
	int size,err;	
	if(filename == NULL)
	{
		printf("Error filename NULL");
		return -1;
	}
	
	FILE* file = fopen(filename,"rb");
	if(file == NULL) 
	{
		printf("Can't open %s\n", filename);
		return -1;
	}
	if((err = fseek(file,0,SEEK_END)) == 0)
	{
		size = ftell(file);
		fseek(file,0,SEEK_SET);
	}
	else
	{
		printf("Error %d %s\n", err, filename);
		return err;
	}
	printf("Size: %d\n", size);

	char* ptr = (char*)malloc(sizeof(char)*size);
	if(ptr != NULL)
	{
		fread(ptr, size, 1, file);
		fclose(file);
	}
	else
	{
		fclose(file);
		return -1;
	}
	
	Convert(filename, ptr, flags);

	free(ptr);
	return 0;
}


void Convert(const char* name, char* buffer, int flags)
{
	bmp_header_t* ptbmp;	
	if((name == NULL)||(buffer == NULL)) return;
	ptbmp = (bmp_header_t*)buffer;

	printf("Bits per pixels: %d\n", ptbmp->bpp);
	printf("Width: %d\n", ptbmp->width);
	printf("Height: %d\n", ptbmp->height);	
	printf("Offset: %d\n", ptbmp->offsetimage);

	if(ptbmp->bpp == 8)
	{
		ConvertToPal8(name,buffer, flags);
	}
	else if(ptbmp->bpp == 24)
	{
		Convert24ToRaw16(name, buffer);
	}
	else
	{
		printf("Unmanaged bit depth: %d\n", ptbmp->bpp);
	}
}

void ConvertToPal8(const char* name, char* buffer, int flags)
{
	bmp_header_t* ptbmp;
	if((name == NULL)||(buffer == NULL)) return;
	ptbmp = (bmp_header_t*)buffer;

	if(flags&REMOVE_PALETTE)
	{
		printf("Convert 8bit bmp file -> raw 8 bits data\n");
	}
	else
	{
		printf("Convert 8bit bmp file -> 16bits 256 colors palette + raw 8 bits data\n");
	}
	int destfilesize = (sizeof(char)*ptbmp->width*ptbmp->height);
	if(!(flags&REMOVE_PALETTE))
	{
		destfilesize += (256*sizeof(unsigned short));
	}

	char* ptpal8 = (char*)malloc(destfilesize);
	if(ptpal8 == NULL)
	{
		printf("Malloc error\n");
		return;
	}
	
	if(!(flags&REMOVE_PALETTE))
	{
		// Converti palette
		printf("\nConverting palette\n");
		unsigned char* pal = (unsigned char*)(buffer + sizeof(bmp_header_t));
		unsigned char* jagpal = (unsigned char*)ptpal8;
		for(int i = 0; i < 256; i++)
		{
			unsigned short col16 = ((pal[2]>>3)<<11) | ((pal[0]>>3)<<6) | (pal[1]>>2);
			*jagpal++ = (col16>>8) & 0xFF; 
			*jagpal++ = col16 & 0xFF;
			pal+=4;
		}
	}

	printf("\nConverting image\n");
	char* dst = ptpal8;
	if(!(flags&REMOVE_PALETTE))
	{
		dst += 256*sizeof(unsigned short);
	}
	char* src = buffer + ptbmp->offsetimage + ptbmp->width*(ptbmp->height-1);
	for(unsigned int j = 0; j < ptbmp->height; j++)
	{
		for(unsigned int i = 0; i <ptbmp->width; i++)
		{
			*dst++ = *src++;
		}
		src -= ptbmp->width*2; // vertical flip
	}

	printf("\nWriting file\n");
	char* destname;
	if(flags&REMOVE_PALETTE)
	{
		destname = DestFileNameRaw8(name);
	}
	else
	{
		destname = DestFileNamePal8(name);
	}
	WriteFile(destname, ptpal8, destfilesize);

	free(ptpal8);
}

void Convert24ToRaw16(const char* name, char* buffer)
{
	bmp_header_t* ptbmp;
	if((name == NULL)||(buffer == NULL)) return;

	printf("Convert 24bits -> Raw 16 bits\n");

	ptbmp = (bmp_header_t*)buffer;

	int destfilesize = sizeof(unsigned short)*ptbmp->width*ptbmp->height;
	unsigned short* ptraw16 = (unsigned short*)malloc(destfilesize);
	if(ptraw16 == NULL)
	{
		printf("Malloc error\n");
		return;
	}
	unsigned char* ptdst = (unsigned char*)ptraw16;
	unsigned char* ptsrc = (unsigned char*)buffer + ptbmp->offsetimage + (ptbmp->width*3)*(ptbmp->height-1);
	for(unsigned int j = 0; j < ptbmp->height; j++)
	{
		for(unsigned int i = 0; i <ptbmp->width; i++)
		{
			int b = (*ptsrc++)>>3;
			int g = (*ptsrc++)>>2;
			int r = (*ptsrc++)>>3;
			//ptsrc++; // skip alpha
			unsigned short rgb16 = (r << 11) | (b<<6) | g;
			*ptdst++ = (rgb16>>8);
			*ptdst++ = rgb16 & 0xFF;
		}
		ptsrc -= ptbmp->width*6; // vertical flip (*3 RGB *2 lignes)
	}

	printf("\nWriting file\n");
	char* destname = DestFileNameRaw16(name);
	WriteFile(destname, (char*)ptraw16, destfilesize);

	free(ptraw16);
}

char* DestFileNamePal8(const char* filename)
{	
	if(filename == NULL) return NULL;
	int len = strlen(filename)+8;	
	char* newname = (char*)malloc(len*sizeof(char));
	if(newname == NULL) return NULL;
	
	strcpy(newname, filename);
	
	char* pext = strrchr(newname, '.');
	if(pext == NULL) return NULL;
	pext++;
	*pext++ = 'p';
	*pext++ = 'a';
	*pext++ = 'l';
	*pext++ = '8';
	*pext++ = 0;

	return newname;
}

char* DestFileNameRaw8(const char* filename)
{	
	if(filename == NULL) return NULL;
	int len = strlen(filename)+8;	
	char* newname = (char*)malloc(len*sizeof(char));
	if(newname == NULL) return NULL;
	
	strcpy(newname, filename);
	
	char* pext = strrchr(newname, '.');
	if(pext == NULL) return NULL;
	pext++;
	*pext++ = 'r';
	*pext++ = 'a';
	*pext++ = 'w';
	*pext++ = '8';
	*pext++ = 0;

	return newname;
}

char* DestFileNameRaw16(const char* filename)
{	
	if(filename == NULL) return NULL;
	int len = strlen(filename)+8;	
	char* newname = (char*)malloc(len*sizeof(char));
	if(newname == NULL) return NULL;
	
	strcpy(newname, filename);
	
	char* pext = strrchr(newname, '.');
	if(pext == NULL) return NULL;
	pext++;
	*pext++ = 'r';
	*pext++ = 'g';
	*pext++ = 'b';
	*pext++ = 0;

	return newname;
}

void WriteFile(const char* filename, char* buffer, int size)
{
	if(filename!=NULL)
	{
		FILE* file = fopen(filename,"wb");
		if(file != NULL)
		{
			fwrite(buffer, size, 1, file);
			fclose(file);
			printf("File %s created\n", filename);
		}
		else
		{
			printf("Error creating %s\n", filename);
		}
		free((void*)filename);
	}
}